/**
 * Copyright (C) 2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.jorm.generator.lib;

import org.apache.velocity.VelocityContext;
import org.apache.velocity.context.Context;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.compiler.api.JormCompilerParameter;
import org.objectweb.jorm.compiler.api.PExceptionCompiler;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.Package;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.metainfo.api.Extension;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.util.io.api.TargetHolder;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.expression.lib.NotEqual;
import org.objectweb.util.monolog.api.BasicLevel;

import java.io.File;
import java.io.FileWriter;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class FPNCGenerator extends CommonGenerator {

	public CommonHelper helper;



	public void generate(Class clazz,
	                     TargetHolder holder,
	                     JormCompilerParameter cp) throws PException {

		Collection nds = clazz.getNameDefs();
		Class classWithNd = clazz;
		while(nds.size() < 1 && classWithNd.getInheritedClassNumber() > 0) {
			classWithNd = ((Extension) classWithNd.getExtensions().iterator().next())
			    .getMOClass();
			nds = classWithNd.getNameDefs();
		}
		if (nds.size() < 1) {
			logger.log(BasicLevel.WARN, "FPNC: No NameDef defined for the class " + clazz.getFQName());
			return;
		}
		if (nds.size() == 1) {
			generate(clazz, (NameDef) nds.iterator().next(), null, holder);
		} else {
			Iterator it = nds.iterator();
			while(it.hasNext()) {
				NameDef nd = (NameDef) it.next();
				String ndName = (nd.getName() != null
				    && nd.getName().length() == 0
				    ? null
				    : nd.getName());
				generate(clazz, nd, ndName, holder);
			}
		}
	}

	/**
	 * This method generates a XFPNC file corresponding to the clazz
	 * parameter in the directory parameter.
	 * @param clazz		The clazz requiring the generation of a PNamingContext.
	 * @param nd		The NameDef to use
	 * @param ndName    The name def name to use fot the package name
	 * (can be null)
	 * @param holder	The target holder which allows to create files.
	 */
	private void generate(Class clazz,
	                      NameDef nd,
	                      String ndName,
	                      TargetHolder holder) throws PException {
		// Calculate fileName
		String packName = ((Package) clazz.getParent()).getName();
		if (ndName != null) {
			if (packName != null && packName.length() > 0) {
				packName += "." + ndName;
			} else {
				packName = ndName;
			}
		}
		String className = clazz.getName() + "FPNC";
		String fileName = className;
		if (packName != null && packName.length() > 0) {
			fileName = packName + "." + fileName;
			fileName = fileName.replace('.', File.separatorChar);
		}
		logger.log(BasicLevel.DEBUG, "Generate the " + fileName);


		Context ctx = new VelocityContext();
		ctx.put("class", clazz);
		ctx.put("helper", this);
		ctx.put("header", GEN_TEMPLATE_DIR + "Header.vm");
		ctx.put("className", className);
		ctx.put("packName", packName);

		if (nd.isNameRef()) {
			ctx.put("isComposite", Boolean.TRUE);
			ctx.put("codingType", "PNameCoder.CTCOMPOSITE");
			ctx.put("compositeName", nd.getNameRef().getCompositeName());
		} else if (nd.isFieldName()) {
			ctx.put("isComposite", Boolean.FALSE);
			TypedElement te = clazz.getTypedElement(nd.getFieldName());
			if (te == null) {
				throw new PException("The field '" + nd.getFieldName()
				    + "' used in the namedef of the class '"
				    + clazz.getFQName() + "' is not defined.");
			} else if (!(te instanceof PrimitiveElement)) {
				throw new PException("MalFormed Meta information: the field '"
				    + nd.getFieldName() + "' used in the namedef of the class '"
				    + clazz.getFQName() + "' is not a primitive element.");
			}
			ctx.put("codingType", getCTDeclaration(te.getType()));
		}

		Expression matchExpression = clazz.getInheritanceFilter(nd);
		MatchInfo mi = new MatchInfo();
		try {
			fillMatchInfo(matchExpression, mi);
		} catch (ExpressionException e) {
			throw new PException(e, "Generation of " + fileName);
		}
		if (mi.filter.length() == 0) {
			mi.filter.append("true");
		} else {
			mi.field2declarations.put("", "PNameGetter png = (PNameGetter) o;");
		}
		ctx.put("mi", mi);

		//Generate the file
		try {
			FileWriter fw = holder.getFileWriter(fileName + ".java");
			velocityEngine.getTemplate(GEN_TEMPLATE_DIR + "FPNC.vm").merge(ctx, fw);
			fw.flush();
			fw.close();
		} catch (Exception e) {
			throw new PExceptionCompiler(e, "Problem while generating " + fileName + ".java");
		}
	}

	public class MatchInfo {
		final static String DEFAULT_PARAM_NAME = "o";
		public Map field2declarations = new HashMap();
		public StringBuffer filter = new StringBuffer();
		public int nbField = 0;
		public PType singleFieldPType = null;
		public String singleFieldName = DEFAULT_PARAM_NAME;

		public Map getField2declarations() {
			return field2declarations;
		}

		public String getFilter() {
			return filter.toString();
		}

		public int getNbField() {
			return nbField;
		}

		public PType getSingleFieldPType() {
			return singleFieldPType;
		}

		public String getSingleFieldName() {
			return singleFieldName;
		}
	}

	public void fillMatchInfo(Expression e,
	                          MatchInfo mi) throws PException, ExpressionException {
		if (e instanceof Operator) {
			Operator operator = (Operator) e;
			logger.log(BasicLevel.DEBUG, "Operator " + operator.getOperatorString());
			int nb = operator.getOperandNumber();
			if (nb == 1) {
				mi.filter.append(operator.getOperatorString());
				fillMatchInfo(operator.getExpression(0), mi);
			} else if (nb == 2) {
				String opPreString = "";
				String opString = operator.getOperatorString();
				String opPostString = "";
				//particular case of the equal operator which is translated
				// - to the equals method on java object,
				// - to a equality (==) on primitive type,
				// - to the static method java.util.Arrays.equals(o1,o2) on arrays
				if (((e instanceof Equal) || (e instanceof NotEqual))
				    && (isObjectOperand(operator.getExpression(0))
				        || isObjectOperand(operator.getExpression(1)))) {
					opString = ".equals(";
					opPostString = ")";
					if (e instanceof NotEqual) {
						opPreString = "!";
					}
					if (isArrayOperand(operator.getExpression(0))
					    || isArrayOperand(operator.getExpression(1))) {
						opPreString += "java.util.Arrays.equals(";
						opString = ",";
					}

				}

				mi.filter.append('(');
				mi.filter.append(opPreString);
				fillMatchInfo(operator.getExpression(0), mi);
				mi.filter.append(opString);
				fillMatchInfo(operator.getExpression(1), mi);
				mi.filter.append(opPostString);
				mi.filter.append(')');
			} else {
				throw new PException("Operators with more than 2 operands are not supported: " + e);
			}
		} else if (e instanceof ParameterOperand) {
			ParameterOperand po = (ParameterOperand) e;
			logger.log(BasicLevel.DEBUG, "ParameterOperand " + po.getName());
			mi.filter.append(po.getName());
			if (mi.field2declarations.get(po.getName()) == null) {
				StringBuffer declaration = new StringBuffer();
				declaration.append(po.getType().getJavaName());
				declaration.append(' ');
				declaration.append(po.getName());
				declaration.append(" = png.pnGet");
				declaration.append(getCoderName(po.getType()));
				declaration.append("Field(");
				declaration.append(po.getName());
				declaration.append(");");
				mi.field2declarations.put(po.getName(), declaration);
				mi.nbField ++;
				mi.singleFieldPType = (mi.nbField == 1
				    ? po.getType()
				    : null);
				mi.singleFieldName = (mi.nbField == 1
				    ? po.getName()
				    : mi.DEFAULT_PARAM_NAME);
				if (logger.isLoggable(BasicLevel.DEBUG)) {
					logger.log(BasicLevel.DEBUG, "Add declaration " + declaration);
					logger.log(BasicLevel.DEBUG, "nbField " + mi.nbField);
					logger.log(BasicLevel.DEBUG, "singleFieldPType " + mi.singleFieldPType);
					logger.log(BasicLevel.DEBUG, "singleFieldName " + mi.singleFieldName);
				}
			}

		} else if ((e instanceof Operand) && ((Operand) e).isDefined()) {
			Operand operand = (Operand) e;
			if (isObjectType(operand.getType())) {
				Object o = operand.getObject();
				if (o == null) {
					mi.filter.append("null");
				}
			} else {
				switch (operand.getType().getTypeCode()) {
				case PType.TYPECODE_CHAR:
				case PType.TYPECODE_OBJCHAR:
					mi.filter.append(operand.getChar());
					break;

				case PType.TYPECODE_BYTE:
				case PType.TYPECODE_OBJBYTE:
					mi.filter.append(operand.getByte());
					break;

				case PType.TYPECODE_SHORT:
				case PType.TYPECODE_OBJSHORT:
					mi.filter.append(operand.getShort());
					break;

				case PType.TYPECODE_INT:
				case PType.TYPECODE_OBJINT:
					mi.filter.append(operand.getInt());
					break;

				case PType.TYPECODE_LONG:
				case PType.TYPECODE_OBJLONG:
					mi.filter.append(operand.getLong());
					break;

				case PType.TYPECODE_STRING:
					mi.filter.append('"');
					mi.filter.append(operand.getString());
					mi.filter.append('"');
					break;

				case PType.TYPECODE_BYTEARRAY: {
					mi.filter.append("new byte[]{");
					String sep = "";
					byte[] arrays = operand.getByteArray();
					for(int i=0; i<arrays.length; i++) {
						mi.filter.append(sep);
						sep = ",";
						mi.filter.append(arrays[i]);
					}
					mi.filter.append("}");
					break;
				}
				case PType.TYPECODE_CHARARRAY: {
					char[] arrays = operand.getCharArray();
						mi.filter.append("new char[]{");
						String sep = "";
						for(int i=0; i<arrays.length; i++) {
							mi.filter.append(sep);
							sep = ",";
							mi.filter.append(arrays[i]);
						}
						mi.filter.append("}");
					break;
				}
				case PType.TYPECODE_DATE:
					mi.filter.append("new java.util.Date(");
					mi.filter.append(operand.getDate().getTime());
					mi.filter.append(')');
					break;
				case PType.TYPECODE_BIGDECIMAL:
					mi.filter.append("new java.math.BigDecimal(");
					mi.filter.append(operand.getBigDecimal().toString());
					mi.filter.append(')');
					break;

				case PType.TYPECODE_BIGINTEGER:
					mi.filter.append("new java.math.BigInteger(");
					mi.filter.append(operand.getBigDecimal().toString());
					mi.filter.append(')');
					break;

				default:

				}
			}
		}
	}
	private boolean isObjectOperand(Expression e) {
		return (e instanceof Operand) && isObjectType(e.getType());
	}

	private boolean isArrayOperand(Expression e) {
		return (e instanceof Operand)
		    && (e.getType().getTypeCode() == PType.TYPECODE_CHARARRAY
		           || e.getType().getTypeCode() == PType.TYPECODE_BYTEARRAY);
	}

}
